import datetime
import glob
import os
import random
import re
import tempfile
import time

import arrow
import numpy as np
import pandas as pd
from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException

from utility_bill_scraper import format_fields


def get_name():
    return "Kitchener-Wilmot Hydro"


def get_electricity_consumption(soup):
    def find_kWh(tag):
        return tag.name == u"span" and (
            tag.decode().find("kWh Off Peak") >= 0
            or tag.decode().find("kWh Mid Peak") >= 0
            or tag.decode().find("kWh On Peak") >= 0
        )

    fields = []
    for x in soup.find_all(find_kWh):
        fields += format_fields(x)

    data = {"off peak": 0, "mid peak": 0, "on peak": 0}

    for x in fields:
        if x.find("kWh Off Peak") > 0:
            data["off peak"] += float(x[: x.find("kWh Off Peak")])
        elif x.find("kWh Mid Peak") > 0:
            data["mid peak"] += float(x[: x.find("kWh Mid Peak")])
        elif x.find("kWh On Peak") > 0:
            data["on peak"] += float(x[: x.find("kWh On Peak")])

    return data


def get_billing_date(soup):
    def find_billing(tag):
        return tag.name == u"div" and tag.decode().find("BILLING DATE") >= 0

    match = re.search(
        "([A-Z]+)\s+(\d+)\s+(\d+)",
        format_fields(
            soup.find_all(find_billing)[0].next_sibling.next_sibling.span.contents
        )[0],
    )
    month, day, year = match.groups()

    return arrow.get("%s %s %s" % (month, day, year), "MMM DD YYYY").date().isoformat()


def get_electricity_rates(soup):
    def find_kWhOffPeak(tag):
        return tag.name == u"div" and tag.decode().find("kWh Off Peak") >= 0

    tag = soup.find(find_kWhOffPeak)
    match = re.search("top:(?P<top>\d+)px", tag.decode())
    top = match.groups()[0]

    # match all divs with the same top pixel coordinate
    def find_matching_top(tag):
        return tag.name == u"div" and tag.decode().find("top:%spx" % top) >= 0

    for x in soup.find_all(find_matching_top):
        fields = format_fields(x.span)
        if len(fields) > 0 and str(fields[0]).find("at $") == 0:
            rates = [float(x[4:]) for x in fields]
            break

    return dict(zip(["off peak", "on peak", "mid peak"], rates))


def get_amount_due(soup):
    def find_new_charges(tag):
        return tag.name == u"div" and tag.decode().find("New Charges") >= 0

    tag = soup.find(find_new_charges)
    match = re.search("top:(?P<top>\d+)px", tag.decode())
    top = float(match.groups()[0])

    def find_left_pos(tag):
        if tag.name == u"div":
            match = re.search("left:(?P<left>\d+)px", tag.decode())
            if match:
                left = float(match.groups()[0])
                return left >= 120 and left <= 132
        return False

    tags = soup.find_all(find_left_pos)

    distance = []
    for tag in tags:
        match = re.search("top:(?P<top>\d+)px", tag.decode())
        distance.append(abs(float(match.groups()[0]) - top))

    amount_due = format_fields(
        tags[np.nonzero(distance == np.min(distance))[0][0]].span
    )[0]

    index = str(amount_due).find("CR")
    if index >= 0:
        amount_due = (-1) * float(amount_due[:index])

    return amount_due


def convert_data_to_df(data):
    df = pd.DataFrame(
        {
            "Date": [arrow.get(x["date"]).date() for x in data],
            "Amount Due": [x["amount due"] for x in data],
            "Off Peak Consumption": [
                x["electricity consumption"]["off peak"] for x in data
            ],
            "Mid Peak Consumption": [
                x["electricity consumption"]["mid peak"] for x in data
            ],
            "On Peak Consumption": [
                x["electricity consumption"]["on peak"] for x in data
            ],
            "Off Peak Rate": [x["electricity rates"]["off peak"] for x in data],
            "Mid Peak Rate": [x["electricity rates"]["mid peak"] for x in data],
            "On Peak Rate": [x["electricity rates"]["on peak"] for x in data],
        }
    )
    df = df.set_index("Date")

    df["Total Consumption"] = (
        df["Off Peak Consumption"]
        + df["Mid Peak Consumption"]
        + df["On Peak Consumption"]
    )
    return df


class Timeout(Exception):
    pass


class KitchenerWilmotHydroAPI:
    name = get_name()

    def __init__(self, user, password, data_directory="."):
        self._user = user
        self._password = password
        self._driver = None
        self._invoice_list = None
        self._temp_download_dir = tempfile.mkdtemp()
        self._data_directory = data_directory
        self._hourly_data_directory = None
        self._invoice_directory = None

    def _init_driver(self, headless=False):
        options = webdriver.ChromeOptions()
        prefs = {
            "download.default_directory": self._temp_download_dir,
            "plugins.always_open_pdf_externally": True,
        }
        options.add_experimental_option("prefs", prefs)
        # options.add_argument('window-size=1200x600')
        options.add_argument("window-size=1920x1080")

        if headless:
            options.add_argument("headless")

        self._driver = webdriver.Chrome(options=options)

    def _close_driver(self):
        self._driver.close()
        self._driver = None

    def _login(self):
        self._driver.get("https://www3.kwhydro.on.ca/app/login.jsp")
        self._driver.find_element_by_id("accessCode").send_keys(self._user)
        self._driver.find_element_by_id("password").send_keys(self._password)
        self._driver.find_element_by_xpath(
            '//*[@id="login-form"]/div[3]/button'
        ).click()

    def _download_link(self, link, ext, timeout=5):
        # remove all files in the temp dir
        files = os.listdir(self._temp_download_dir)
        for file in files:
            os.remove(os.path.join(self._temp_download_dir, file))

        # add a random delay to keep from being banned
        time.sleep(1 * random.random() * 3)

        # download the link
        link.click()

        t_start = time.time()
        filepath = None
        # wait for the file to finish downloading
        while time.time() - t_start < timeout:
            files = glob.glob(os.path.join(self._temp_download_dir, "*.%s" % ext))
            if len(files):
                filepath = os.path.join(self._temp_download_dir, files[0])
                time.sleep(0.5)
                break
        if not filepath:
            raise Timeout
        return filepath

    def download_hourly_data(self, start_date=None, end_date=None):
        self._init_driver(headless=False)

        try:
            self._login()

            yesterday = arrow.get().date() - datetime.timedelta(days=1)

            if not start_date:
                start_date = "%d-%02d-01" % (yesterday.year - 2, yesterday.month)
            start_date = arrow.get(start_date).date()

            if end_date:
                end_date = arrow.get(end_date).date()
            else:
                end_date = arrow.get().date()

            if end_date > start_date and end_date > yesterday:
                date_range = pd.date_range(start_date, yesterday, freq="M").append(
                    pd.DatetimeIndex([yesterday])
                )
            else:
                date_range = pd.date_range(start_date, end_date, freq="M")
            date_range = [x.date() for x in date_range]

            self._hourly_data_directory = os.path.join(
                self._data_directory, self.name, "hourly data"
            )

            for date in date_range:
                if not os.path.isdir(self._hourly_data_directory):
                    os.makedirs(self._hourly_data_directory)

                new_filepath = os.path.join(
                    self._hourly_data_directory, "%s.csv" % (date.isoformat())
                )

                if os.path.exists(new_filepath):
                    continue

                print(
                    "Downloading hourly data for %d-%02d-01 to %d-%02d-%02d..."
                    % (date.year, date.month, date.year, date.month, date.day)
                )

                # Wait a random period between requests so that we don't get blocked
                time.sleep(5 + random.random() * 5)

                url = (
                    "https://www3.kwhydro.on.ca/app/capricorn?para=smartMeterConsum&inquiryType=hydro"
                    "&fromYear=%d&fromMonth=%02d&fromDay=%02d&toYear=%d&toMonth=%02d&toDay=%02d"
                    % (date.year, date.month, 1, date.year, date.month, date.day)
                )
                self._driver.get(url)

                def is_valid_date_range():
                    valid_date_range = True
                    try:
                        valid_date_range = not self._driver.find_element_by_class_name(
                            "alert.alert-danger"
                        ).text.startswith(
                            "You are not authorized to view the selected date range."
                        )
                    except NoSuchElementException:
                        pass
                    return valid_date_range

                # Check if this date range is valid
                if not is_valid_date_range():
                    print("  No data for this date range.")
                    continue

                # Wait a random period between requests so that we don't get blocked
                time.sleep(5 + random.random() * 5)

                filepath = self._download_link(
                    self._driver.find_element_by_id("download"), "csv"
                )

                # Read the csv file
                df_csv = pd.read_csv(filepath)[
                    :-1
                ]  # Ignore last line which is just a note

                # Create an hourly index
                index = pd.date_range(
                    df_csv["Reading Date"].iloc[0],
                    (
                        arrow.get(df_csv["Reading Date"].iloc[-1])
                        + datetime.timedelta(days=1)
                    )
                    .date()
                    .isoformat(),
                    freq="h",
                )[:-1]

                # Create a new dataframe to store the reformatted data
                df = pd.DataFrame({"kWh": np.zeros(len(index))}, index=index)

                # Reformat the data indexed by a timestamp
                for i, row in df_csv.iterrows():
                    df.loc[row[0], "kWh"] = df_csv.loc[i][1:25].values

                # Offset index by 1hr
                df.index = [x + datetime.timedelta(hours=1) for x in df.index]

                # If we're downloading data from the current month, delete
                # previous files
                if date == yesterday:
                    files = glob.glob(
                        os.path.join(
                            self._hourly_data_directory,
                            "%d-%02d-*.csv" % (date.year, date.month),
                        )
                    )
                    for f in files:
                        os.remove(f)

                # Save the reformatted data
                df.to_csv(new_filepath)
        finally:
            self._close_driver()

    def download_invoices(self, start_date=None, end_date=None, timeout=5):
        if self._invoice_list is not None:
            return self._invoice_list

        self._init_driver(headless=False)

        self._invoice_directory = os.path.abspath(
            os.path.join(self._data_directory, self.name, "invoices")
        )
        if not os.path.isdir(self._invoice_directory):
            os.makedirs(self._invoice_directory)

        try:
            self._login()

            # convert start and end dates to date objects
            if start_date:
                start_date = arrow.get(start_date).date()
            if end_date:
                end_date = arrow.get(end_date).date()

            # Iterate through the invoices in reverse chronological order
            # (i.e., newest invoices are first).

            # Open Bills & Payment
            link = self._driver.find_element_by_link_text("Bills & Payment")
            self._driver.execute_script("arguments[0].scrollIntoView(true);", link)
            time.sleep(0.5)
            link.click()

            self._driver.switch_to.frame("iframe-BILLINQ")
            time.sleep(0.5)
            bills_table = self._driver.find_element_by_id("billsTable")

            rows = [
                [y for y in x.find_elements_by_tag_name("td")]
                for x in bills_table.find_element_by_tag_name(
                    "tbody"
                ).find_elements_by_tag_name("tr")[2:]
            ]

            data = []
            for row in rows:
                row_data = [x.text for x in row[1:]]

                date = arrow.get(row_data[0], "MMM D, YYYY").date()

                if start_date and date < start_date:
                    break

                if end_date and date > end_date:
                    continue

                data.append(row_data)
                new_filepath = os.path.join(
                    self._invoice_directory,
                    "%s - %s - $%s.pdf"
                    % (date.isoformat(), self.name, row_data[1].split(" ")[1]),
                )

                def download_link(link, ext, timeout=5):
                    # remove all files in the temp dir
                    files = os.listdir(self._temp_download_dir)
                    for file in files:
                        os.remove(os.path.join(self._temp_download_dir, file))

                    # add a random delay to keep from being banned
                    time.sleep(1 * random.random() * 3)

                    # download the link
                    link.click()

                    t_start = time.time()
                    filepath = None
                    # wait for the file to finish downloading
                    while time.time() - t_start < timeout:
                        files = glob.glob(
                            os.path.join(self._temp_download_dir, "*.%s" % ext)
                        )
                        if len(files):
                            filepath = os.path.join(self._temp_download_dir, files[0])
                            time.sleep(0.5)
                            break
                    if not filepath:
                        raise Timeout
                    return filepath

                if not os.path.exists(new_filepath):
                    # download the pdf invoice
                    img = row[0].find_element_by_tag_name("img")
                    filepath = download_link(img, "pdf")
                    os.rename(filepath, new_filepath)
        finally:
            self._close_driver()

        # Convert to a dictionary
        results = {}
        for i, name in enumerate(["Invoice Date", "Invoice Amount"]):
            results[name] = [x[i] for x in data]

        # Reformat dates
        for name in ["Invoice Date"]:
            results[name] = [
                arrow.get(x, "MMM D, YYYY").date().isoformat() for x in results[name]
            ]

        self._invoice_list = pd.DataFrame(results)

        return self._invoice_list



import re

import arrow
import pandas as pd

from utility_bill_scraper import format_fields


def get_name():
    return "Enbridge"


def get_bill_date(soup):
    def find_bill_date(tag):
        return tag.name == u"div" and tag.decode().find("Bill Date") >= 0

    return format_fields(soup.find(find_bill_date).contents[1].contents)[0]


def get_amount_due(soup):
    pos_re = (
        "left:(?P<left>\d+)px.*top:(?P<top>\d+)px.*"
        "width:(?P<width>\d+)px.*height:(?P<height>\d+)"
    )

    def find_amount_due_now(tag):
        return tag.name == u"div" and tag.decode().find("Amount due now") >= 0

    tag = soup.find_all(find_amount_due_now)[-1]
    pos = re.search(pos_re, tag.decode()).groupdict()
    pos = {k: int(v) for (k, v) in pos.items()}
    pos["bottom"] = pos["top"] + pos["height"]
    pos["right"] = pos["left"] + pos["width"]

    def find_divs_on_same_line(tag):
        if tag.name == u"div":
            match = re.search(pos_re, tag.decode())
            if match:
                top = int(match.groupdict()["top"])
                bottom = top + int(match.groupdict()["height"])
                left = int(match.groupdict()["left"])
                left + int(match.groupdict()["width"])
                return (left > pos["right"]) and (
                    (top >= pos["top"] and top <= pos["bottom"])
                    or (bottom >= pos["top"] and bottom <= pos["bottom"])
                )
        return False

    return format_fields(soup.find(find_divs_on_same_line).span.contents)[0][1:]


def get_summary(soup):
    def find_gas_used_this_period(tag):
        return tag.name == u"div" and tag.decode().find("Gas used this period") >= 0

    div = soup.find(find_gas_used_this_period)

    field_data = format_fields(div.next_sibling.span.contents)

    """
    field_names = [format_fields(x.contents) for x in div.contents]

    # Flatten the list of lists.
    field_names = [item for sublist in field_names for item in sublist]
    """

    # Dynamic discovery of field names failing. Hard-code for now.
    field_names = [
        u"Meter Number",
        u"Estimated Reading",
        u"Previous Reading",
        u"Gas used this period",
        u"PEF Value",
        u"Adjusted volume",
    ]

    summary_dict = dict(zip(field_names, field_data))
    summary_dict[u"Bill Date"] = get_bill_date(soup)
    summary_dict[u"Amount Due"] = get_amount_due(soup)

    return summary_dict


def convert_data_to_df(data):
    cols = data[0]["summary"].keys()
    data_sets = []
    for col in cols:
        data_sets.append([x["summary"][col] for x in data])
    df = pd.DataFrame(data=dict(zip(cols, data_sets)))

    df["Bill Date"] = [arrow.get(x, "MMM DD, YYYY").date() for x in df["Bill Date"]]
    df = df.set_index("Bill Date")

    return df



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['utility_bill_scraper', 'utility_bill_scraper.bin']

package_data = \
{'': ['*']}

install_requires = \
['arrow>=1.2.0,<2.0.0',
 'beautifulsoup4>=4.10.0,<5.0.0',
 'google-api-python-client>=2.27.0,<3.0.0',
 'matplotlib>=3.4.3,<4.0.0',
 'numpy>=1.21.2,<2.0.0',
 'pandas>=1.3.3,<2.0.0',
 'pdfminer>=20191125,<20191126',
 'selenium>=3.141.0,<4.0.0']

setup_kwargs = {
    'name': 'utility-bill-scraper',
    'version': '0.3.6',
    'description': 'Utility bill scraper for extracting data from pdfs and websites.',
    'long_description': '# Utility bill scraper\n\n[![build](https://github.com/ryanfobel/utility-bill-scraper/actions/workflows/build.yml/badge.svg?branch=main)](https://github.com/ryanfobel/utility-bill-scraper/actions/workflows/build.yml)\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/ryanfobel/utility-bill-scraper/main)\n[![PyPI version shields.io](https://img.shields.io/pypi/v/utility-bill-scraper.svg)](https://pypi.python.org/pypi/utility-bill-scraper/)\n\nExtract energy usage and carbon footprint from utility websites or pdf bills. Currently, this library supports:\n\n * [Kitchener Utilities (gas & water)](https://www.kitchenerutilities.ca)\n * [Kitchener-Wilmot Hydro (electricity)](https://www.kwhydro.on.ca)\n\n# Install\n\n```\npip install utility-bill-scraper\n```',
    'author': 'Ryan Fobel',
    'author_email': 'ryan@fobel.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import calendar
import glob
import os
import random
import re
import shutil
import tempfile
import time

import arrow
import numpy as np
import pandas as pd
from selenium.common.exceptions import (
    NoSuchElementException,
    StaleElementReferenceException,
)

from utility_bill_scraper import Timeout, UtilityAPI, convert_divs_to_df, format_fields


def get_name():
    return "Kitchener Utilities"


def get_summary(soup):
    def find_seq_id(tag):
        return tag.name == "div" and tag.decode().find("SEQ-ID") >= 0

    def find_account_summary(tag):
        return tag.name == "span" and tag.decode().find("Your Account Summary") >= 0

    summary_fields = format_fields(soup.find_all(find_account_summary)[0].contents)
    summary_data = format_fields(
        soup.find_all(find_seq_id)[0].next_sibling.contents[0].contents
    )

    summary_dict = dict(zip(summary_fields[1:], summary_data))

    def find_charges(name):
        def find_matching_div(tag):
            return tag.name == "div" and tag.decode().find(name) >= 0

        tag = soup.find(find_matching_div)

        # Extract the top pixel coordinate.
        match = re.search(r"top:(?P<top>\d+)px", tag.decode())
        top = match.groups()[0]

        # Find the second div with the same top pixel coordinate.
        return format_fields(
            soup.find_all(style=re.compile("top:%spx" % top))[1].span.contents
        )[0]

    summary_dict["Water Charges"] = find_charges("Water charges")
    summary_dict["Gas Charges"] = find_charges("Gas charges")

    return summary_dict


def get_water_consumption(soup):
    def find_total_consumption(tag):
        return tag.name == "div" and tag.decode().find("Total Consumption") >= 0

    div_list = soup.find_all(find_total_consumption)

    # Find the div containing 3 fields (gas has an extra
    # 'Billing Conversion Multiplier'). Note that it is possible to have
    # more than one consumption section.

    tags = [x for x in div_list if len(format_fields(x.contents[0])) == 3]

    consumption = []

    for tag in tags:
        # Extract the top pixel coordinate.
        match = re.search(r"top:(?P<top>\d+)px", tag.decode())
        top = match.groups()[0]

        # Match all divs with the same top pixel coordinate.
        def find_matching_top(tag):
            return tag.name == "div" and tag.decode().find("top:%spx" % top) >= 0

        divs = [format_fields(x.contents[0]) for x in soup.find_all(find_matching_top)]
        consumption.append(dict(zip(divs[0], divs[2])))
    return consumption


def get_water_and_sewer_charges(soup):
    def find_water_consumption(tag):
        return (
            (tag.name == "div")
            and (tag.decode().find("Consumption") >= 0)
            and (tag.decode().find("Total Consumption") == -1)
        )

    water_div = soup.find_all(find_water_consumption)[0]
    water_type = format_fields(water_div.next_sibling.contents[0])

    result = {"Time period": water_type[0]}
    water_type = water_type[1:]

    consumption = format_fields(water_div.next_sibling.next_sibling.contents[0])
    rates = format_fields(
        water_div.next_sibling.next_sibling.next_sibling.next_sibling.contents[0]
    )
    charges = format_fields(
        water_div.next_sibling.next_sibling.next_sibling.next_sibling.next_sibling.contents[
            0
        ]
    )

    for x in range(len(water_type)):
        result[water_type[x]] = {
            "Consumption": consumption[x],
            "Rate": rates[x],
            "Charges": charges[x],
        }
    return result


def get_water_charges(soup):
    data = get_water_and_sewer_charges(soup)
    types = ["Water", "Sewer"]
    return dict(zip(types, [data[x]["Charges"] for x in types]))


def get_water_rates(soup):
    data = get_water_and_sewer_charges(soup)
    types = ["Water", "Sewer"]
    return dict(zip(types, [data[x]["Rate"] for x in types]))


def get_gas_consumption(soup):
    def find_total_consumption(tag):
        return tag.name == "div" and tag.decode().find("Total Consumption") >= 0

    div_list = soup.find_all(find_total_consumption)

    # Find divs containing 4 fields (gas has an extra
    # 'Billing Conversion Multiplier'). Note that it is possible to have
    # more than one consumption section.
    tags = [x for x in div_list if len(format_fields(x.contents[0])) > 3]

    consumption = []
    for tag in tags:
        # Extract the top pixel coordinate.
        match = re.search(r"top:(?P<top>\d+)px", tag.decode())
        top = match.groups()[0]

        # Match all divs with the same top pixel coordinate.
        def find_matching_top(tag):
            return tag.name == "div" and tag.decode().find("top:%spx" % top) >= 0

        divs = [format_fields(x.contents[0]) for x in soup.find_all(find_matching_top)]

        consumption.append(dict(zip(divs[0], divs[2])))
    return consumption


def get_gas_charges(soup):
    try:
        # Find the bounding box that defines the gas section.
        pos_re = r"left:(?P<left>\d+)px.*top:(?P<top>\d+)px.*width:(?P<width>\d+)px.*height:(?P<height>\d+)"

        def find_gas_section(tag):
            return tag.name == "div" and tag.decode().find("GAS") >= 0

        tag = soup.find(find_gas_section)
        pos = re.search(pos_re, tag.decode()).groupdict()
        top_bound = int(pos["top"])

        def find_gas_charges(tag):
            return tag.name == "div" and tag.decode().find("Gas charges") >= 0

        tag = soup.find(find_gas_charges)
        pos = re.search(pos_re, tag.decode()).groupdict()
        bottom_bound = int(pos["top"])

        # Find all of the div tags within this bounding box.
        def find_divs_within_bounds(tag):
            match = re.search(pos_re, tag.decode())
            if match:
                top = int(match.groupdict()["top"])
                return top >= top_bound and top < bottom_bound and tag.name == "div"
            return False

        df = convert_divs_to_df(soup.find_all(find_divs_within_bounds))
        df["fields_str"] = [str(x) for x in df["fields"]]
        df = df.sort_values(["top", "left"])

        # Charges can be grouped in different sections (e.g., if the gas rate
        # changes in the middle of the month). We only care about the last
        # section, because it contains the Fixed Delivery Charge.
        charges = df[
            df["left"] > df[df["fields_str"] == "[u'Charges']"]["left"].iloc[0]
        ].iloc[-1]["fields"]
        charge_desc = df[df["fields_str"].str.find(" days") >= 0].iloc[-1]["fields"][1:]
        return dict(zip(charge_desc[-len(charges) :], charges))
    except AttributeError as error:
        print("Error scraping gas charges")
    return {}


def get_gas_rates(soup):
    def find_gas_rates(tag):
        return tag.name == "div" and tag.decode().find("Gas Fixed Delivery Charge") >= 0

    gas_div = soup.find_all(find_gas_rates)[0]
    gas_fields = format_fields(gas_div.contents[0])
    gas_fields = gas_fields[1:]
    gas_rates = format_fields(
        gas_div.next_sibling.next_sibling.next_sibling.contents[0]
    )
    format_fields(
        gas_div.next_sibling.next_sibling.next_sibling.next_sibling.contents[0]
    )

    return dict(
        zip(
            [
                x + " Rate"
                for x in gas_fields
                if (x.find("HST") == -1) and x.find("Fixed") == -1
            ],
            gas_rates,
        )
    )


class KitchenerUtilitiesAPI(UtilityAPI):
    name = get_name()

    def __init__(
        self,
        user=None,
        password=None,
        data_path=None,
        file_ext=".csv",
        headless=True,
        timeout=10,
        save_statements=True,
        google_sa_credentials=None,
    ):
        super().__init__(
            user=user,
            password=password,
            data_path=data_path,
            file_ext=file_ext,
            headless=headless,
            timeout=timeout,
            save_statements=save_statements,
            google_sa_credentials=google_sa_credentials,
        )

    def _login(self):
        self._driver.get(
            "https://ebilling.kitchener.ca/sap/bc/ui5_ui5/sap/ZUMCUI5/index.html"
        )
        self._driver.find_element_by_id("CANCEL_BUTTON ").click()
        self._driver.find_element_by_id("__field1").send_keys(self._user)
        self._driver.find_element_by_id("__field0").send_keys(self._password)
        self._driver.find_element_by_id("__button0").click()

    def _get_header_nav_bar(self):
        result = None
        t_start = time.time()
        while time.time() - t_start < self._timeout:
            try:
                pages = self._driver.find_element_by_id(
                    "headerNavigationBar"
                ).find_elements_by_tag_name("li")
                keys = [x.text for x in pages]
                result = dict(zip(keys, pages))
                result.pop("", None)
                return result
            except NoSuchElementException:
                pass
        raise Timeout

    def _get_contracts(self):
        # Pick the account (e.g., "Gas", "Water and Sewer", "Stormwater")
        contract_table = self._driver.find_element_by_id(
            "ContractTable-table"
        ).find_element_by_tag_name("tbody")
        rows = contract_table.find_elements_by_tag_name("tr")

        contracts = [x.find_elements_by_tag_name("td")[0] for x in rows]
        keys = [x.text for x in contracts]
        return dict(zip(keys, contracts))

    def _first_page(self):
        # Get a list of the pages available
        t_start = time.time()
        link = None
        while time.time() - t_start < self._timeout:
            try:
                link = self._driver.find_element_by_id(
                    "__table1-paginator--firstPageLink"
                )
                link.location_once_scrolled_into_view
                link.click()
                return
            except NoSuchElementException:
                pass
            except StaleElementReferenceException:
                pass
        raise Timeout

    def _get_pages(self):
        # Get a list of the pages available
        table = None
        t_start = time.time()
        while time.time() - t_start < self._timeout:
            try:
                table = self._driver.find_element_by_id("__table1-paginator-pages")
                time.sleep(1)
                break
            except NoSuchElementException:
                pass

        if table:
            time.sleep(0.5)
            pages = table.find_elements_by_tag_name("li")
            keys = [int(x.text) for x in pages]
            return dict(zip(keys, pages))
        else:
            raise Timeout
        return None

    def download_statements(self, start_date=None, end_date=None, max_downloads=None):
        download_path = tempfile.mkdtemp()
        self._init_driver()
        downloaded_files = []

        try:
            self._login()

            # convert start and end dates to date objects
            if start_date:
                start_date = arrow.get(start_date).date()
            if end_date:
                end_date = arrow.get(end_date).date()

            def get_data():
                # Iterate through the statements in reverse chronological order
                # (i.e., newest statements are first).

                billing_table = self._driver.find_element_by_id("__table1-table")

                rows = [
                    [y for y in x.find_elements_by_tag_name("td")]
                    for x in billing_table.find_element_by_tag_name(
                        "tbody"
                    ).find_elements_by_tag_name("tr")
                ]

                data = []
                for row in rows:
                    row_data = [x.text for x in row[1:]]
                    if row_data[0] == "":
                        continue

                    date = arrow.get(row_data[1], "MM/DD/YYYY").date()

                    if start_date and date < start_date:
                        break

                    if end_date and date > end_date:
                        continue

                    if max_downloads and len(downloaded_files) >= max_downloads:
                        break

                    data.append(row_data)
                    new_filepath = os.path.join(
                        download_path,
                        "%s - %s - $%s.pdf"
                        % (date.isoformat(), self.name, row_data[3]),
                    )

                    def download_link(link, ext):
                        # remove all files in the temp dir
                        files = os.listdir(self._temp_download_dir)
                        for file in files:
                            os.remove(os.path.join(self._temp_download_dir, file))

                        # add a random delay to keep from being banned
                        time.sleep(1 * random.random() * 3)

                        # download the link
                        link.click()

                        t_start = time.time()
                        filepath = None
                        # wait for the file to finish downloading
                        while time.time() - t_start < self._timeout:
                            files = glob.glob(
                                os.path.join(self._temp_download_dir, "*.%s" % ext)
                            )
                            if len(files):
                                filepath = os.path.join(
                                    self._temp_download_dir, files[0]
                                )
                                time.sleep(0.5)
                                break
                        if not filepath:
                            raise Timeout
                        return filepath

                    downloaded_files.append(new_filepath)
                    if not os.path.exists(new_filepath):
                        # download the pdf statement
                        for img in row[0].find_elements_by_tag_name("img"):
                            if img.get_property("title") == "PDF":
                                filepath = download_link(img, "pdf")
                                shutil.move(filepath, new_filepath)

                return data

            self._get_header_nav_bar()["BILLING"].click()
            self._first_page()

            pages_downloaded = []
            pages_to_download = list(self._get_pages().keys())
            data = []
            while len(pages_to_download):
                pages = self._get_pages()

                # Add any new pages that are not in our list
                for x in pages.keys():
                    if x not in pages_to_download and x not in pages_downloaded:
                        pages_to_download += [x]

                # Download the first page in the list
                i = pages_to_download.pop(0)

                pages[i].click()
                pages_downloaded += [i]
                time.sleep(1)
                page_data = get_data()
                # If this page returned some data within the date range, add it
                if len(page_data):
                    data += page_data
                # If there was no data within the date range and we alread have data,
                # it means that any remaining pages must be prior to the start date;
                # therefore we should stop.
                elif len(data):
                    break
        finally:
            self._close_driver()

        return downloaded_files

    def get_consumption_history(self, contract):
        self._init_driver()

        try:

            def get_data():
                # The Consumption history div (this contains all of the data we are interested in)
                consumption_history = (
                    self._driver.find_element_by_id("contractConsumptionHistory")
                    .find_element_by_id("__table1-table")
                    .find_element_by_tag_name("tbody")
                )
                rows = consumption_history.find_elements_by_tag_name("tr")
                data = [
                    [y.text for y in x.find_elements_by_tag_name("td")] for x in rows
                ]
                return data

            self._login()

            self._get_header_nav_bar()["ACCOUNTS"].click()
            self._get_contracts()[contract].click()

            # Click on the "Consumption History" tab
            link = self._driver.find_element_by_id("contractDetailNavigationBarItem2")
            link.location_once_scrolled_into_view
            link.click()
            time.sleep(0.5)

            data = []
            for page in self._get_pages().keys():
                link = self._get_pages()[page]
                link.location_once_scrolled_into_view
                link.click()
                data += get_data()

            # sum values that have the same date
            dates = [x for x, y in data]
            results = dict(zip(dates, [0] * len(dates)))
            results.pop("", None)
            for date, consumption in data:
                if date != "":
                    results[date] += float(consumption)
        finally:
            self._close_driver()

        def convert_to_series(data):
            dates = [
                arrow.get("%s 01 %s" % tuple(x.split(" ")), "MMMM DD YYYY").date()
                for x in list(data.keys())
            ]
            dates = [
                "%d-%02d-%02d"
                % (x.year, x.month, calendar.monthrange(x.year, x.month)[1])
                for x in dates
            ]
            return pd.Series(list(data.values()), index=dates)

        return convert_to_series(results)

    def convert_data_to_df(self, data):
        cols = list(data[0]["summary"].keys())
        if "Pre-authorized Withdrawal" in cols:
            cols.remove("Pre-authorized Withdrawal")
            cols.append("Total Due")

        for x in data:
            if "Pre-authorized Withdrawal" in x["summary"].keys():
                x["summary"]["Total Due"] = x["summary"]["Pre-authorized Withdrawal"]
        data_sets = []
        for col in cols:
            data_sets.append([x["summary"][col] for x in data])

        df = pd.DataFrame(data=dict(zip(cols, data_sets)))

        df["Issue Date"] = [
            str(arrow.get(x, "MMM DD YYYY").date()) for x in df["Issue Date"]
        ]
        df = df.set_index("Issue Date")

        # Extract water and gas consumption.
        water_consumption = [
            np.sum(
                [
                    x["Total Consumption"] if "Total Consumption" in x else 0
                    for x in row["water consumption"]
                ]
            )
            for row in data
        ]
        gas_consumption = [
            np.sum(
                [
                    x["Total Consumption"] if "Total Consumption" in x else 0
                    for x in row["gas consumption"]
                ]
            )
            for row in data
        ]
        """
        # To do: gas charges were broken when updating to python3
        
        gas_fixed_charge = [x['gas charges']['Gas Fixed Delivery Charge']
                            if 'Gas Fixed Delivery Charge'
                            in x['gas charges'] else None for x in data]
        # Figure out the sales tax rate.
        sales_tax_on_gas = [x['gas charges']['HST on Gas']
                            if 'HST on Gas'
                            in x['gas charges'] else None for x in data]
        sales_tax_rate = df['Gas Charges'] / (
            df['Gas Charges'] - sales_tax_on_gas) - 1

        # Adjust the fixed charge by the sales tax
        df['Gas Fixed Charges'] = np.round(gas_fixed_charge * (
            1 + sales_tax_rate), 2)

        # Calculate the variable charge and rate.
        df['Gas Variable Charges'] = df['Gas Charges'] - df['Gas Fixed Charges']
        df['Gas Variable Rate'] = (df['Gas Variable Charges'] /
                                df['Gas Consumption'])
        """

        df["Water Consumption"] = water_consumption
        df["Gas Consumption"] = gas_consumption

        return df



import datetime as dt
import glob
import io
import json
import os
import re
import shutil
import subprocess
import tempfile

import arrow
import pandas as pd
from apiclient import discovery
from bs4 import BeautifulSoup
from google.oauth2.service_account import Credentials
from googleapiclient.http import MediaFileUpload, MediaIoBaseDownload
from selenium import webdriver

DEFAULT_SCOPES = [
    "https://www.googleapis.com/auth/spreadsheets",
    "https://www.googleapis.com/auth/drive",
]

# Natural gas emission factor
# 119.58 lbs CO2/1000 cubic feet of natural gas
# 1.915 kg CO2/1 m^3 natural gas [119.58 lbs * (1 kg / 2.204623 lbs) *
#   (1 ft^3 / (0.0254 * 12)**3 m^3) / 1000]
GAS_KGCO2_PER_CUBIC_METER = (
    119.58 * (1 / 2.204623) * (1 / (0.0254 * 12) ** 3) / 1000
)  # kg CO2/1 m^3 natural gas

# Natural gas energy density
# 1,037 Btu / ft^3 (https://www.eia.gov/tools/faqs/faq.php?id=45&t=8)
# Energy per m^3: 1,037 Btu / ft^3 * 1055.1 J / 1 Btu * 1 ft^3 /
#   (0.0254 * 12)**3 m^3
#   37 MJ/m3 (https://hypertextbook.com/facts/2002/JanyTran.shtml)

GAS_JOULES_PER_CUBIC_METER = 1037 * 1055.1 / (0.0254 * 12) ** 3  # J / m^3
KWH_PER_JOULE = 1.0 / (60 * 60 * 1000)
GAS_KWH_PER_CUBIC_METER = GAS_JOULES_PER_CUBIC_METER * KWH_PER_JOULE


def is_number(s):
    """Returns True if string is a number."""
    return s.replace(".", "", 1).isdigit()


def format_fields(input_list):
    """Strip newlines and trailing colons from fields and convert numbers to
    floats."""
    fields = [x.strip().strip(":") for x in input_list if x.find("</br>") == -1]
    return [float(x) if is_number(x) else x for x in fields]


def convert_divs_to_df(divs):
    """Convert list of divs to a pandas DataFrame describing the position and
    geomtery of each tag."""
    pos_re = (
        r"left:(?P<left>\d+)px.*top:(?P<top>\d+)px.*"
        r"width:(?P<width>\d+)px.*height:(?P<height>\d+)"
    )

    df = pd.DataFrame()
    for x in divs:
        pos = re.search(pos_re, x.decode()).groupdict()
        try:
            df = df.append(
                pd.DataFrame(
                    dict(
                        left=int(pos["left"]),
                        top=int(pos["top"]),
                        width=int(pos["width"]),
                        height=int(pos["height"]),
                        fields=[format_fields(x.span.contents)],
                    )
                )
            )
        except AttributeError:
            pass

    df["right"] = df["left"] + df["width"]
    df["bottom"] = df["top"] + df["height"]
    return df


def is_kitchener_utilities_bill(soup):
    """Check if this is a Kitchener Utilities bill."""

    def find_kitchener_utilities(tag):
        return (
            tag.name == "span"
            and tag.decode().find("Supplier: KITCHENER UTILITIES") >= 0
        )

    return len(soup.find_all(find_kitchener_utilities)) > 0


def is_kitchener_wilmot_hydro_bill(soup):
    """Check if this is a Kitchener-Wilmot Hydro bill."""

    def find_kitchener_wilmot_hydro(tag):
        return (
            tag.name == "span" and tag.decode().find("KITCHENER-WILMOT HYDRO INC") >= 0
        )

    return len(soup.find_all(find_kitchener_wilmot_hydro)) > 0


def is_enbridge_gas_bill(soup):
    """Check if this is an Enbridge Gas bill."""

    def find_enbridge_gas(tag):
        return (
            tag.name == "span"
            and tag.decode().find("Enbridge Gas Distribution Inc.") >= 0
        )

    return len(soup.find_all(find_enbridge_gas)) > 0


def process_pdf(pdf_file, rename=False, keep_html=False):
    """Extract data from a pdf file and return a nested Python dictionary.
    Optionally rename the pdf with the format:
        YYYY-MM-DD-UTILITY_NAME-$XX.xx.pdf
    If the `keep_html` flag is True, the intermediate html fill will be
    maintained."""
    basename, ext = os.path.splitext(pdf_file)
    basepath = os.path.dirname(pdf_file)
    html_file = basename + ".html"

    if os.path.exists(os.path.join(os.getenv("CONDA_PREFIX"), "Scripts", "pdf2txt.py")):
        subprocess.check_output(
            [
                "python",
                r"%CONDA_PREFIX%\Scripts\pdf2txt.py",
                "-o%s" % html_file,
                pdf_file,
            ],
            shell=True,
        )
    else:
        subprocess.check_output(
            ["pdf2txt.py '-o%s' '%s'" % (html_file, pdf_file)],
            shell=True,
        )

    with open(html_file, "r", encoding="utf8") as f:
        soup = BeautifulSoup(f, "html.parser")

    if keep_html:
        # re-write formatted html (useful for debugging)
        with open(html_file, "w", encoding="utf8") as f:
            f.write(str(soup.prettify().encode("utf-8")))
    else:
        os.remove(html_file)

    result = None
    new_name = None

    if is_kitchener_utilities_bill(soup):
        from . import kitchener_utilities as ku

        summary = ku.get_summary(soup)

        if "Pre-authorized Withdrawal" in summary.keys():
            amount_due = summary["Pre-authorized Withdrawal"]
        elif "Total Due" in summary.keys():
            amount_due = summary["Total Due"]
        else:
            print("Couldn't find amount due!")
            amount_due = None

        new_name = "%s - %s - $%.2f.pdf" % (
            arrow.get(summary["Issue Date"], "MMM DD YYYY").format("YYYY-MM-DD"),
            ku.get_name(),
            amount_due,
        )

        # To do: several functions were broken when updating to python3
        result = {
            "name": ku.get_name(),
            "summary": summary,
            "water consumption": ku.get_water_consumption(soup),
            # 'water and sewer charges': ku.get_water_and_sewer_charges(soup),
            "gas consumption": ku.get_gas_consumption(soup),
            # 'gas charges': ku.get_gas_charges(soup),
            # 'gas rates': ku.get_gas_rates(soup)
        }
    elif is_enbridge_gas_bill(soup):
        from . import enbridge as en

        summary = en.get_summary(soup)

        new_name = "%s - %s - $%s.pdf" % (
            arrow.get(summary["Bill Date"], "MMM DD, YYYY").format("YYYY-MM-DD"),
            en.get_name(),
            summary["Amount Due"],
        )

        result = {"name": en.get_name(), "summary": summary}
    elif is_kitchener_wilmot_hydro_bill(soup):
        from . import kitchener_wilmot_hydro as kwh

        date = kwh.get_billing_date(soup)
        amount_due = kwh.get_amount_due(soup)
        new_name = "%s - %s - $%.2f.pdf" % (date, kwh.get_name(), amount_due)

        result = {
            "name": kwh.get_name(),
            "date": date,
            "amount due": amount_due,
            "electricity rates": kwh.get_electricity_rates(soup),
            "electricity consumption": kwh.get_electricity_consumption(soup),
        }
    else:
        print("Unrecognized bill type.")

    if rename and new_name:
        os.rename(pdf_file, os.path.join(basepath, new_name))

    return result


def convert_data_to_df(data):
    """Convert the list of dictionaries (returned from `process_pdf`) into a
    dictionary of pandas DataFrames keyed by utility name."""
    result = {}

    for x in data:
        if x["name"] not in result.keys():
            result[x["name"]] = pd.DataFrame()
        if x["name"] == "Kitchener Utilities":
            from . import kitchener_utilities as ku

            df = ku.convert_data_to_df([x])
        elif x["name"] == "Enbridge":
            from . import enbridge as en

            df = en.convert_data_to_df([x])
        elif x["name"] == "Kitchener-Wilmot Hydro":
            from . import kitchener_wilmot_hydro as kwh

            df = kwh.convert_data_to_df([x])
        else:
            print("Unknown name.")
            continue
        result[x["name"]] = result[x["name"]].append(df)

    return result


def is_gdrive_path(path):
    if path:
        return path.startswith("https://drive.google.com/drive")
    else:
        return False


class Timeout(Exception):
    pass


class UnsupportedFileTye(Exception):
    pass


class GDriveHelper:
    def __init__(self, service_account_info):
        if type(service_account_info) == str:
            service_account_info = json.loads(service_account_info)
        self._credentials = Credentials.from_service_account_info(
            info=service_account_info,
            scopes=DEFAULT_SCOPES,
        )
        self._service = discovery.build("drive", "v3", credentials=self._credentials)

    def get_file_in_folder(self, folder_id, file_name):
        # Query the shared google folder for file that matches `file_name`
        return (
            self._service.files()
            .list(q=f"'{folder_id}' in parents and name='{file_name}'")
            .execute()["files"][0]
        )

    def get_file(self, file_id):
        # Query google drive for a file matching the `file_id`
        return self._service.files().get(fileId=file_id).execute()

    def create_subfolder(self, parent_folder_id, name):
        file_metadata = {
            "name": name,
            "mimeType": "application/vnd.google-apps.folder",
            "parents": [parent_folder_id],
        }
        file = self._service.files().create(body=file_metadata, fields="id").execute()
        return file

    def create_file_in_folder(self, folder_id, local_path):
        print(
            f"Upload file to google drive folder(folder_id={folder_id}, local_path={local_path}"
        )
        file_metadata = {"name": os.path.basename(local_path), "parents": [folder_id]}
        media = MediaFileUpload(local_path, mimetype="text/csv", resumable=True)
        file = (
            self._service.files()
            .create(body=file_metadata, media_body=media, fields="id")
            .execute()
        )
        return file

    def upload_file(self, file_id, local_path):
        print(f"Upload file to google drive(file_id={file_id}, local_path={local_path}")
        file = self.get_file(file_id)
        media_body = MediaFileUpload(
            local_path, mimetype=file["mimeType"], resumable=True
        )
        updated_file = (
            self._service.files()
            .update(fileId=file["id"], media_body=media_body)
            .execute()
        )
        return updated_file

    def download_file(self, file_id, local_path):
        print(
            f"Download file from google drive(file_id={file_id}, local_path={local_path}"
        )
        file = self._service.files().get_media(fileId=file_id)
        fh = io.BytesIO()
        downloader = MediaIoBaseDownload(fh, file)
        done = False
        while done is False:
            status, done = downloader.next_chunk()

        # Make parent dirs if necessary.
        os.makedirs(os.path.split(local_path)[0], exist_ok=True)

        # Write the file.
        with open(local_path, "wb") as f:
            f.write(fh.getbuffer().tobytes())


class UtilityAPI:
    def __init__(
        self,
        user=None,
        password=None,
        data_path=None,
        file_ext=".csv",
        headless=True,
        timeout=10,
        save_statements=True,
        google_sa_credentials=None,
    ):
        self._user = user
        self._password = password
        self._driver = None
        self._browser = None
        self._headless = headless
        self._temp_download_dir = tempfile.mkdtemp()
        self._data_path = data_path or os.path.abspath(os.path.join(".", "data"))
        self._file_ext = file_ext
        self._save_statements = save_statements

        if google_sa_credentials is not None:
            self._gdh = GDriveHelper(google_sa_credentials)
        else:
            self._gdh = None

        supported_filetypes = [".csv"]
        if self._file_ext not in supported_filetypes:
            raise UnsupportedFileTye(
                f"`file_ext`={self._file_ext} has an invalid filetype. Acceptable extensions are "
                f'{",".join([f"`{x}`" for x in supported_filetypes])}'
            )

        self._history = pd.DataFrame()

        # If `data_path` is a google drive folder download the data file.
        if is_gdrive_path(self._data_path):
            if google_sa_credentials is None:
                raise RuntimeError(
                    "`data_path` looks like a google drive folder, but `google_sa_credentials` is None."
                )

            folder_id = self._data_path.split("/")[-1]
            utility_folder = self._gdh.get_file_in_folder(folder_id, self.name)
            data_file = self._gdh.get_file_in_folder(
                utility_folder["id"], "data" + self._file_ext
            )
            self._gdh.download_file(
                data_file["id"],
                os.path.join(self._temp_download_dir, "data" + self._file_ext),
            )
            self._history = pd.read_csv(
                os.path.join(self._temp_download_dir, "data" + self._file_ext)
            ).set_index("Issue Date")
        elif os.path.exists(
            os.path.join(self._data_path, self.name, "data" + self._file_ext)
        ):
            # Load csv with previously cached data if it exists locally.
            self._history = pd.read_csv(
                os.path.join(self._data_path, self.name, "data" + self._file_ext)
            ).set_index("Issue Date")

        self._timeout = timeout

    def _init_driver(self, browser="Firefox"):
        self._browser = browser
        if self._browser == "Chrome":
            options = webdriver.ChromeOptions()
            prefs = {"download.default_directory": self._temp_download_dir}
            options.add_experimental_option("prefs", prefs)
            if self._headless:
                options.add_argument("--window-size=1920,1080")
                options.add_argument("--headless")
                options.add_argument("--disable-gpu")
                options.add_argument(
                    "user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36"
                )
            self._driver = webdriver.Chrome(options=options)
        elif self._browser == "Firefox":
            options = webdriver.firefox.options.Options()
            options.set_preference("browser.download.folderList", 2)
            options.set_preference("browser.download.dir", self._temp_download_dir)
            options.set_preference("browser.download.useDownloadDir", True)
            options.set_preference(
                "browser.download.viewableInternally.enabledTypes", ""
            )
            options.set_preference(
                "browser.helperApps.neverAsk.saveToDisk",
                "application/pdf;text/plain;application/text;text/xml;application/xml",
            )
            options.set_preference("pdfjs.disabled", True)
            # disable the built-in PDF viewer
            if self._headless:
                options.add_argument("--headless")
                options.add_argument("--window-size=1920,1080")
                options.add_argument("--start-maximized")
                options.add_argument("--disable-gpu")
                options.add_argument("--no-sandbox")
            self._driver = webdriver.Firefox(options=options)

    def history(self):
        return self._history

    def __del__(self):
        if self._driver:
            self._close_driver()

    def _close_driver(self):
        self._driver.close()
        self._driver = None

    def update(self, max_downloads=None):
        # Download any new statements.
        start_date = None
        if len(self._history):
            start_date = (
                arrow.get(self._history.index[-1]).date() + dt.timedelta(days=1)
            ).isoformat()
        pdf_files = self.download_statements(
            start_date=start_date, max_downloads=max_downloads
        )

        if self._save_statements:
            # If `data_path` is a google drive url, upload pdfs to gdrive.
            if is_gdrive_path(self._data_path):
                folder_id = self._data_path.split("/")[-1]
                try:
                    utility_folder = self._gdh.get_file_in_folder(folder_id, self.name)
                except IndexError:
                    utility_folder = self._gdh.create_subfolder(folder_id, self.name)
                try:
                    statements_folder = self._gdh.get_file_in_folder(
                        utility_folder["id"], "statements"
                    )
                except IndexError:
                    statements_folder = self._gdh.create_subfolder(
                        utility_folder["id"], "statements"
                    )
                for local_path in pdf_files:
                    self._gdh.create_file_in_folder(statements_folder["id"], local_path)
            else:
                # If `data_path` is a local path, copy pdfs to their new location.
                os.makedirs(
                    os.path.join(self._data_path, self.name, "statements"),
                    exist_ok=True,
                )
                for local_path in pdf_files:
                    shutil.move(
                        local_path,
                        os.path.join(
                            self._data_path,
                            self.name,
                            "statements",
                            os.path.basename(local_path),
                        ),
                    )

                # Update the `pdf_files` list
                pdf_files = [
                    os.path.join(
                        self._data_path,
                        self.name,
                        "statements",
                        os.path.basename(local_path),
                    )
                    for local_path in pdf_files
                ]
        return self._scrape_pdf_files(pdf_files)

    def _scrape_pdf_files(self, pdf_files=[]):
        cached_invoice_dates = list(self._history.index)

        # Scrape data from pdf files
        data = []
        for pdf_file in pdf_files:
            date = os.path.splitext(os.path.basename(pdf_file))[0].split(" - ")[0]

            # If we've already scraped this pdf, continue
            if date not in cached_invoice_dates:
                print("Scrape data from %s" % pdf_file)
                try:
                    result = process_pdf(pdf_file, rename=True)
                    if result:
                        data.append(result)
                except Exception as e:
                    print(e)

        # Convert the to a dataframe
        if len(data):
            df = self.convert_data_to_df(data)
            self._history = self._history.append(df)

            # Update history

            # If `data_path` is a google drive folder, upload the data file.
            if is_gdrive_path(self._data_path):
                folder_id = self._data_path.split("/")[-1]
                try:
                    utility_folder = self._gdh.get_file_in_folder(folder_id, self.name)
                except IndexError:
                    utility_folder = self._gdh.create_subfolder(folder_id, self.name)

                self._history.to_csv(
                    os.path.join(self._temp_download_dir, "data" + self._file_ext)
                )

                try:
                    data_file = self._gdh.get_file_in_folder(
                        utility_folder["id"], "data" + self._file_ext
                    )
                    self._gdh.upload_file(
                        data_file["id"],
                        os.path.join(self._temp_download_dir, "data" + self._file_ext),
                    )
                except IndexError:
                    data_file = self._gdh.create_file_in_folder(
                        utility_folder["id"],
                        os.path.join(self._temp_download_dir, "data" + self._file_ext),
                    )

            else:
                # Create directories if necessary
                os.makedirs(os.path.join(self._data_path, self.name), exist_ok=True)

                # Update csv file
                self._history.to_csv(
                    os.path.join(self._data_path, self.name, "data" + self._file_ext)
                )

            return df
